# /tcl/km-categories.tcl
#
# Categorization of knowledge objects.
#
# Created by carsten@arsdigita.com in June 2000
#
# Modified by tvujec@arsdigita.com in August 2001
#
# $Id: km-categories-procs.tcl,v 1.2 2009/02/13 20:40:39 jeffd Exp $

# Category hierarchy
#
#     location <question_name> [Placeholder]
#       Global <real category> [Location]

ad_proc km_category_table {
    {
	-object_type_id 0
	-question_id 0
	-parent_node_id 0
        -uncategorized_p 1 
        -target {}
        -category_counts {}
	-maximum_sub_columns 1
    }
} {

    Generate the table of links for a category tree. Returns 0 if
    the table is empty, otherwise returns a non-zero integer.

} {
    template::multirow create categorization_widget category target category_id sum_count parent_node_id question_id
    template::multirow create categorization_widget_header header_name
    if { !$parent_node_id } {

	# If no parent categories have been specified, we get all the
	# categories that can be associated with the given object type
	# and question.

	set questions [db_list_of_lists get_category_questions {
	    select q.question_id, q.node_id, q.pretty_name
	    from sn_questions q, sn_question_object_type_map qm
	    where qm.object_type_id = :object_type_id
		and qm.question_id = q.question_id
		and qm.question_state <> 'invisible'
	        and q.browse_p = 't'
		and q.abstract_data_type in ('category','other_category')
	    order by qm.sort_key}]

	if [null_p $questions] { return 0 }
    } else {
	set pretty_question [db_string get_pretty_question {
	    select pretty_name
	    from sn_questions
	    where question_id = :question_id}]
	set questions [list [list $question_id $parent_node_id $pretty_question]]
    }

    set category_list {}
    foreach question $questions {
	# For each of the parent categories we display the child categories.

	set question_id [fst $question]
	set parent_node_id [snd $question]
	set pretty_question [thd $question]

	if { $uncategorized_p && ![empty_string_p $category_counts] } {
	    set uncategorized_target "${target}none&question_id=$question_id"
	    set uncategorized_count [ns_set get $category_counts "uncat $question_id"]
	} else {
	    set uncategorized_target ""
	    set uncategorized_count 0
	}

	set elem [km8_categorization_widget \
		-uncategorized_p $uncategorized_p \
		-parent_node_id $parent_node_id -target $target \
		-question_id $question_id \
		-category_count $category_counts \
		-uncategorized_target $uncategorized_target \
		-uncategorized_count $uncategorized_count \
		-maximum_columns $maximum_sub_columns]

	if ![null_p $elem] {
	    lappend category_list [list $pretty_question $elem]
	}
    }

    set n_total 0
    set i 0
    set category_type_count [llength $category_list]
    while { $i < $category_type_count } {
	template::multirow append categorization_widget_header [fst [lindex $category_list $i]]
	incr i
    }

    return $i
}

ad_proc km_get_child_categories {question_id} {

    Returns all the categories associated with the given question
    as a list of tuples holding the category_id and the indented
    category name.

} {
    set root_node_id [db_string get_node_id {
	select node_id from sn_questions 
	where question_id = :question_id
    }]

    if ![empty_string_p $root_node_id] {
	set result [util_memoize "swc_get_category_tree $root_node_id 0 {}"]
    } else {
	set result {}
    }

    return $result
}

ad_proc km_get_node_id {question_id} {

    Returns the node_id for this question_id or an empty string.

} {
    set result [db_string km_get_category_id "
	select node_id from sn_questions where question_id=:question_id" -default ""]
    return $result
}

ad_proc km_categorize_object { question_id object_id category_ids } {

    Inserts a row into the sw_object_category_map which maps each
    object to a list of categories.

    Also takes care of removing old categorizations which are of the same type
    as the given category_ids.

    Returns a list of 1 for each successfully inserted and 0 for each
    failed category_id.

} {
    foreach category_id $category_ids {
	if ![empty_string_p $category_id] {
	    lappend nonempty_category_ids $category_id
	}
    }
    if {![info exists nonempty_category_ids]} {
	km_uncategorize_object $object_id $question_id
	return
    }
    set category_ids $nonempty_category_ids

    set user_id [ad_conn user_id]
    set ip_address [ns_conn peeraddr]

    db_transaction {
        # First get rid of any old categorizations of this object
	# that belongs to the same question.

        db_dml uncategorize {
	    delete from sw_object_category_map
	    where object_id = :object_id
		and category_id in (select child_category_id from sw_flat_cat 
				    where parent = (select node_id from sn_questions
						    where question_id = :question_id))
	}

        # Insert the new categorization.
        # The insert query gracefully ignores duplicate categorization.

        foreach category_id $category_ids {
	    db_dml categorize {
		insert into sw_object_category_map (object_id, category_id) 
		      select :object_id, :category_id
		      from dual
		      where not exists (select 1 from sw_object_category_map
		                        where object_id = :object_id
		                        and category_id = :category_id)
	    }
        }

        db_exec_plsql last_modified {
            begin
            acs_object.last_modified (
            v_object_id           => :object_id,
            v_modifying_user      => :user_id,
            v_modifying_ip        => :ip_address
            );
            end;
        }

        set pretty_categories [db_list km_categorize_object_50 "
	    select long_name from categories 
	    where category_id in ([join $category_ids ","])"]

        if { [llength $pretty_categories] > 1 } {
            set label "categories"
        } else {
            set label "category"
        }

	set content "Set $label [DoubleApos [join $pretty_categories ", "]]."
        db_exec_plsql km_categorize_object_6 {
            begin
            object.audit_object (
            v_object_id              => :object_id,
            v_question_id            => :question_id,
            v_last_modifying_user_id => :user_id,
	    v_modifying_ip           => :ip_address,
            v_content                => :content
            );
            end;
        }
    }

    # After the user has changed the categorization of an object, we
    # want to give him up-to-date category counts. So we flush the
    # memoized km_get_category_count results for him.
    km_flush_category_count -object_id $object_id
}

ad_proc km_flush_category_count {
    {
	-object_id 0
	-object_type_id 0
	-user_id 0
	-local 0
    }
} {
    Clears the cached category count for the current user.
    Note that we only flush for "begin_letter all". Flushing
    for all letters would take too much time.
} {
    if !$user_id { set user_id [ad_conn user_id] }
    if !$object_type_id { set object_type_id [km_get_object_type $object_id] }

    if !$local {
	server_cluster_httpget_from_peers "/SYSTEM/flush-category-count.tcl?[export_url_vars object_type_id user_id]"
    }

    foreach pool {all own coown} {
	foreach age {90 180 365 730 all} {
	    foreach archived {t f} {
		util_memoize_flush_local "km_get_category_count_internal -age $age -pool $pool -begin_letter all -archived $archived $object_type_id $user_id"
	    }
	}
    }
}

ad_proc km_uncategorize_object { object_id question_id } {

    Removes all categorization of this object for this question.
    This is used for non-mandatory fields were the user hasn't selected 
    any categories.

} {
    db_transaction {
        db_dml uncategorize {
	    delete from sw_object_category_map
	    where object_id = :object_id
		and category_id in (select child_category_id from sw_flat_cat
				    where parent = (select node_id from sn_questions 
						    where question_id = :question_id))
	}

        set user_id [ad_conn user_id]
        set ip_address [ns_conn peeraddr]

        db_exec_plsql km_uncategorize_object_10 {
            begin
            acs_object.last_modified (
            v_object_id           => :object_id,
            v_modifying_user      => :user_id,
            v_modifying_ip        => :ip_address
            );
            end;
        }

	set content "Deleted all categories."
	set ip_address [ns_conn peeraddr]
        db_exec_plsql km_uncategorize_object_3 {
            begin
            object.audit_object (
            v_object_id              => :object_id,
            v_question_id            => :question_id,
            v_last_modifying_user_id => :user_id,
	    v_modifying_ip           => :ip_address,
            v_content                => :content
            );
            end;
        }
    }

    km_flush_category_count -object_id $object_id
}

ad_proc km_get_categories {object_id question_id } {
    Returns a list of {category_id category} mapped to this object and this question.
} {
    return [db_list_of_lists get_categories {
	select c.category_id, c.long_name 
	from categories c, sw_flat_cat fc, sw_object_category_map cm, sn_questions q
        where cm.object_id = :object_id
	    and q.question_id = :question_id
	    and fc.parent = q.node_id
	    and c.category_id = fc.child_category_id
	    and cm.category_id = c.category_id
    }]
}

ad_proc km_get_category_counts {
    {
	-age "all"
	-pool "all"
	-begin_letter "all"
	-start_category_id 0
	-archived "f"
	-start_date ""
	-end_date ""
    }
    object_type_id result 
} {

    Returns the count of number of objects as a new ns_set. 
    Caches things for a few minutes.

} {
    set category_count_timeout [ad_parameter CategoryCountTimeout library 300]
    set user_id [ad_conn user_id]

    set category_count_list [util_memoize "km_get_category_count_internal -age $age -pool $pool -begin_letter $begin_letter -start_category_id $start_category_id -archived $archived -start_date $start_date -end_date $end_date $object_type_id $user_id" $category_count_timeout]

    foreach elem $category_count_list {
	ns_set put $result [lindex $elem 0] [lindex $elem 1]
    }

    return $result
}

ad_proc km_get_category_count_internal {
    {
	-age "all"
	-pool "all"
	-begin_letter "all"
	-start_category_id 0
	-archived "f"
	-start_date ""
	-end_date ""
    }
    object_type_id user_id
} {

    Counts the number of objects the user can view, grouped by
    category. (Internal version.)

} {
    ! We're not caching.  Entering km_get_category_count_internal
    set package_id [ad_conn package_id]
    set instance_read_p [ad_permission_p $package_id "read"]

    set category_count_list {}

    lappend where_clauses \
	"o.object_type_id = :object_type_id" \
	"o.context_id = :package_id" \
	"o.expiration_date > sysdate" \
	"o.archived_p = :archived"

    set tables "sw_category_dim swcd, sw_flat_cat fc, sw_object_category_map swcm, sn_objects o"
    set total_tables "sn_objects o"
    if { !$instance_read_p && $pool == "all" } {
	# Only count published objects, or objects for which the user
	# has view permission.

	# Note that we only check direct read permission on flat
	# groups and users. Since the context_id of private objects is
	# NULL, and we only grant permissions to private groups, this
	# is sufficient.

        lappend where_clauses \
            "(o.public_p = 't' or o.object_id in (
              select object_id from acs_permissions
              where (grantee_id = :user_id
                     or grantee_id in (select group_id from membership_rels
                                       where person_id = :user_id))
                  and privilege = 'read'))"
    }

    if { $age != "all" } { lappend where_clauses "sysdate - o.last_modified < :age" }

    # Filter by author.
    switch $pool {
	own {
	    lappend where_clauses "o.original_author_id = :user_id"
	}
	coown {
	    set tables "(select group_id as grantee_id from membership_rels where person_id = :user_id union select $user_id as grantee_id from dual) g, acs_permissions p, sn_objects o, sw_object_category_map swcm, sw_flat_cat fc, sw_category_dim swcd"
	    set total_tables "(select group_id as grantee_id from membership_rels where person_id = :user_id union select $user_id as grantee_id from dual) g, acs_permissions p, sn_objects o"
	    lappend where_clauses \
		    "o.object_id = p.object_id" \
		    "p.grantee_id = g.grantee_id" \
		    "p.privilege = 'write'"
	}
    }

    if { ![string equal $begin_letter all] } {
	if { ![string equal $begin_letter other] } {
	    lappend where_clauses "upper(o.one_line_description) like '[string toupper $begin_letter]%'"
	} else {
	    lappend where_clauses "(upper(o.one_line_description) < 'A' or upper(o.one_line_description) > 'Z')"
	}
    }

    if { ![empty_string_p $start_date] && ![empty_string_p $end_date]} {
	lappend where_clauses "o.start_date <= :end_date" "o.end_date >= :start_date"
    }

    # Count the total number of viewable objects.

    set total_viewable_objects_count [db_string get_total_count [subst {
	select count(unique o.object_id)
	from $total_tables
	where [join $where_clauses "\nand "]
    }]]

    lappend category_count_list [list total $total_viewable_objects_count]

    if { $start_category_id == 0 } {

	# Count the uncategorized stuff for each question.
	# We do this by counting all the categorized objects
	# and subtracting the resulting number from the total number 
	# of objects. This is much faster than letting Oracle do
	# the job.

	set questions [km_node_question_list -browse_p $object_type_id]

	foreach question $questions {
	    util_unlist $question question_node_id dummy question_id

	    set count [db_string count_categorized_objects [subst {
		select /*+ORDERED*/ count(unique o.object_id)
		from $total_tables, sw_object_category_map swcm, sw_flat_cat fc
		where [join $where_clauses "\n and "]
		    and swcm.object_id = o.object_id
		    and swcm.category_id = fc.child_category_id
		    and fc.parent = :question_node_id
	    }]]

	    lappend category_count_list \
		[list "uncat $question_id" [expr $total_viewable_objects_count - $count]]

	    # Count categorized stuff by category on the first level below the question.

	    db_foreach get_category_counts [subst {
		select /*+ORDERED*/ fc.parent_category_id, 
		    count(unique o.object_id) as object_count
		from $tables
		where [join $where_clauses "\n and "]
		    and swcd.parent_node_id = :question_node_id
		    and fc.parent = swcd.node_id
		    and swcm.category_id = fc.child_category_id
		    and swcm.object_id = o.object_id
		group by fc.parent_category_id
	    }] {
		lappend category_count_list [list $parent_category_id $object_count]
	    }
	}

    } else {

	# Get counts below a certain root category.

	set count [db_string count_categorized_objects [subst {
	    select /*+ORDERED*/ count(unique o.object_id)
	    from $total_tables, sw_object_category_map swcm, sw_flat_cat fc
	    where [join $where_clauses "\n and "]
		and swcm.object_id = o.object_id
		and swcm.category_id = fc.child_category_id
		and fc.parent_category_id = :start_category_id
	}]]

	lappend category_count_list \
	    [list $start_category_id $count] \
	    [list "uncat" [expr $total_viewable_objects_count - $count]]

	# Count categorized stuff by category on the first level below the question.

	set node_id [db_string get_start_node_id {
	    select node_id 
	    from sw_category_dim 
	    where object_id = :start_category_id
	}]

	db_foreach get_category_counts [subst {
	    select /*+ORDERED*/ fc.parent_category_id, 
		count(unique o.object_id) as object_count
	    from $tables
	    where [join $where_clauses "\n and "]
		and swcd.parent_node_id = :node_id
		and fc.parent = swcd.node_id
		and swcm.category_id = fc.child_category_id
		and swcm.object_id = o.object_id
	    group by fc.parent_category_id
	}] {
	    lappend category_count_list [list $parent_category_id $object_count]
	}

    }

    ! Exiting km_get_category_count_internal
    return $category_count_list
}

ad_proc km_node_question_list { {-browse_p:boolean 0} object_type_id } {

    Returns a list of questions that are categorizations for the given
    object type, each element being a list of 
    node ID, category or tree name, question ID, pretty_name.
    if -browse_p is given, it will return only browseable category
    questions.

} {
    if {$browse_p} {
	set browse_clause "and q.browse_p = 't'"
    } else {
	set browse_clause ""
    }

    set result [db_list_of_lists km_node_question_list "
        select q.node_id, g.tree_name, q.question_id, q.pretty_name
        from sn_questions q, sn_question_object_type_map qm, generic_trees g
        where q.question_id = qm.question_id
	    and g.tree_id = q.tree_id
	    and qm.object_type_id = :object_type_id
	    and qm.question_state <> 'invisible'
            $browse_clause
	order by qm.sort_key"]

    return $result
}

ad_proc km_question_in_object_type { } {
    Replaces the PL/SQL function sn_question_in_object_type. 
    Returns an exists clause that should be much faster than the previous
PL/SQL function.  
} {
 return "and 1= (select count(*) 
	from sn_question_object_type_map
	where object_type_id = :object_type_id 
	    and question_id = :question_id
	    and question_state <> 'invisible')"
}

ad_proc km_node_question { object_type_id node_id } {

    Given an object type and a top-level node ID, determine 
    which question this category belongs to.

} {
    # rownum=1 added by dirk (11-july-2001). remove it...it's most likely
    # wrong, but I just want to get this big piece of software to run

    set result [db_string km_node_question "
	select question_id from sn_questions
	where node_id=:node_id
	    [km_question_in_object_type] and rownum=1" -default ""]
    return $result
}

ad_proc km_node_question_parent { object_type_id node_id } {

    Given an object type and a node ID, determine which question this
    category belongs to.

} {
    return [km_db_1row_to_list km_node_question_parent {
	select q.question_id, q.node_id
	from sn_questions q, sw_flat_cat fc, sn_question_object_type_map qm
	where fc.child = :node_id
	    and fc.parent = q.node_id
	    and qm.object_type_id = :object_type_id
	    and qm.question_id = q.question_id
	    and qm.question_state <> 'invisible'
    }]
}

ad_proc km_category_name { category_id } {
    Get the name of the of a category
} {
    if [empty_string_p $category_id]  { return -code error "Missing category_id" }
    if {[string equal $category_id none]} {
	return ""
    }
    set result [db_string km_category_name "
	select long_name from categories 
	where category_id = :category_id" -default ""]

    return $result
}

ad_proc km_root_node { 
    {
	-question_id 0
    } 
    object_type_id
} {

    Get the parent node_id for a given object type.

    If question_id is 0, a list of the parents for all the questions
    of this object type is returned.
    Else, the parent for the given question_id is returned.

} {
    if !$question_id {
	# Select all categorization questions.

	set result [db_list km_root_node "
	select node_id
	from sn_questions q, sn_question_object_type_map qm
	where q.node_id is not null
	    and qm.object_type_id = :object_type_id
	    and qm.question_id = q.question_id
	    and qm.question_state <> 'invisible'
	order by sort_key"]
    } else {
	set result [db_string km_root_node_2 "
	select node_id from sn_questions where question_id=:question_id" -default ""]
    }

    return $result
}	

ad_proc km_object_filter_bar { 
    {
	-child_type_p 0
    }
    pool age sort_by 
} {

    Filters and sorting options displayed on the top of most category pages.
    Note that *you* have to provide the TABLE tag.

} {
    return "
<tr><td>Owner:</td><td>[km_ownership_bar $pool]</td></tr>
<tr><td>Sort: <td>[km_sort_by_bar $sort_by $child_type_p]</td></tr>"
}

ad_proc km_object_view_bar_top_level { show short_title all_title } {
    View: Just category descriptions | All objects 
} {
    set show_list [list [list short $short_title] [list all $all_title]]

    foreach show_pair $show_list {
	set show_value [lindex $show_pair 0]
	set show_annotation [lindex $show_pair 1]

	if {$show == $show_value} {
	    lappend show_bar "<b>$show_annotation</b>"
	} else {
	    lappend show_bar "<a href=\"[ns_conn url]?[export_ns_set_vars url {show}]&show=$show_value\">$show_annotation</a>"
	}
    }

    return "\[ [join $show_bar " | "] \]"
}    

ad_proc km_ownership_bar { current_pool } {
    Filter: My objects | My objects & co-owner | All
} {
    return [ad_choice_bar \
	{ "my objects" "my objects & co-owner" "all" } \
	"[ns_conn url]?[export_ns_set_vars url {pool}]&pool=\$value" \
	{ own coown all } \
	$current_pool]
}

ad_proc km_archive_bar { current_archive } {
    Filter: Live Objects | Archived Objects
} {
    return [ad_choice_bar \
	{ "live objects" "archived objects" } \
	"[ns_conn url]?[export_ns_set_vars url {archive}]&archive=\$value" \
	{ f t } \
	$current_archive]
}

ad_proc km_age_bar { current_age } {
    Filter: 3 6 12 24 months | All
} {
    proc mk_link { default parms } {
	set label [fst $parms]
	set value [snd $parms]

	if { $value == $default } {
	    return "<strong>$label</strong>"
	} else {
	    return "<a href=\"[ns_conn url]?[export_ns_set_vars url {age}]&age=$value\">$label</a>"
	}
    }

    return "[join [map "mk_link $current_age" [list "3 90" "6 180" "12 365" "24 730"]] " "] months | [mk_link $current_age "All all"]"
}

ad_proc km8_categorization_widget {
    {
        -parent_node_id {}
        -uncategorized_p 0
        -target {}
	-question_id {}
        -category_count {}
	-uncategorized_target {}
	-uncategorized_count {}
	-maximum_columns 1
    }
} {

    Build the table showing category counts for a given parent category.

} {
    # Translate ns_set category_count into a Tcl array
    #
    if { ![null_p $category_count] } { 
        for {set i 0} {$i < [ns_set size $category_count]} {incr i} { 
            set sum_count([ns_set key $category_count $i])  [ns_set value $category_count $i]
        }
    }

    # Fetch the category hierarchy.
    #

    set ordered {}
    set select_size 0
    db_foreach categories {
	select c.long_name as category, c.category_id
	from categories c, sw_category_dim d
	where d.parent_node_id = :parent_node_id
	    and c.category_id = d.object_id
	order by nvl(c.sort_key,9999)
    } {
        if { [empty_string_p $category_count] ||
	     ([info exists sum_count($category_id)] && $sum_count($category_id) > 0) } {
	    lappend ordered [list $category $category_id]
	    incr select_size
	}
    }

    if { !$select_size } { return "" }

    # Generate the widget open tag, now that we know the size
    if { $maximum_columns > 1 } {
	if { $select_size < 15 } {
	    set maximum_columns 1
	} else {
	    set column_size [expr $select_size / $maximum_columns]
	}
    }

    # What column are we currently writing to
    set column_no 1
    # How many rows have we written to that
    set column_count 0
    set show_category_p 0

    # Iterate to generate the HTML
    #
    foreach elem $ordered {
	set category [fst $elem]
	set category_id [snd $elem]

	set option ""

	if { $maximum_columns > 1 } {
	    incr column_count
	    if { $column_count > $column_size && $column_no < $maximum_columns } {
		incr column_no
		set column_count 0
	    }
	}

	if {![info exists sum_count($category_id)]} {
	    set sum_count($category_id) ""
	}
		
	set show_category_p 1
	template::multirow append categorization_widget $category $target $category_id "$sum_count($category_id)" $parent_node_id $question_id

    } 
    # foreach 
	
    if { $uncategorized_p } {
	if {![empty_string_p $uncategorized_count] && $uncategorized_count != 0} {
	    if { ![empty_string_p $uncategorized_target]} {
		set target ${uncategorized_target}
	    }
	    set show_category_p 1
	    template::multirow append categorization_widget Uncategorized $target "" $uncategorized_count $parent_node_id

	}
    }
    return $show_category_p 
}


ad_proc ad_category_parentage_list {category_id {known_parent_id ""}} {

    Returns a list of lists, where each sublist is one line of
    parentage up from the specified category to the hierarchy root. In
    turn, each parentage line list consists of two-item lists: category_id
    and long_name. A list of lists is needed since a category can have
    multiple parents. If this category has no parents, then return the
    empty list.

} {

    set n_parents [db_string ad_category_parentage_list_10 "select count(*) from sw_category_dim where object_id = :category_id"]

    if { $n_parents == 0 } {
	return [list]
    }

    set sql_ad_category_parentage_list_20  "SELECT c.category_id AS parent_id, c.long_name AS parent_category, hc.level_col
FROM categories c,
(SELECT h.object_id as child, LEVEL AS level_col, ROWNUM AS row_col
 FROM sw_category_dim h
 START WITH h.object_id = :category_id
 CONNECT BY h.node_id = PRIOR h.parent_node_id) hc
WHERE c.category_id = hc.child
ORDER BY hc.row_col"

    set parentage_list [list]
    set parentage_line [list]
    set prior_level 0
    set stop_everything 0
    db_foreach ad_category_parentage_list_20 $sql_ad_category_parentage_list_20 {
        if {$stop_everything== 1} {

            if {$level_col <= $prior_level} {
                # Parent line is now completed, flush it
                lappend parentage_list $parentage_line
                
                # Keep the last bit of the parentage chain
                # lars@pinds.com, Feb 2000
                #
                set parentage_line [lrange $parentage_line [expr $prior_level - $level_col + 1] end]
            }
            
            set prior_level $level_col
            
            # We're moving up the hierarchy so put this category at
            # the beginning of the parentage line.
            #
            set parentage_line [concat [list [list $parent_id $parent_category]] $parentage_line]
            
            if { ![empty_string_p $known_parent_id] && $known_parent_id == $parent_id } {
                set stop_everything 1
            }   
        }

    }

    if {$stop_everything == 1} {
        # you cannot return from within db_foreach
        return $parentage_line
    }
    if { [empty_string_p $known_parent_id] } {
	# Don't forget the last parentage line
	return [lappend parentage_list $parentage_line]
    } else {
	# We didn't find the wanted parent
	return [list]
    }
}
