# /tcl/km-links.tcl
#
# Knowledge object linking.
#
# Created by carsten@arsdigita.com in June 2000
#
# $Id: km-links-procs.tcl,v 1.2 2009/02/13 20:40:39 jeffd Exp $

ad_proc km_linked_object_types { object_type_id } {

    Returns an array in the form: object_type_id, pretty_type, pretty_plural, graphic 
    for all object_types that are linkable with this object_type.

} {
    # Only object_types that have been listed for a question in
    # sn_questions are considered to be linkable with this
    # object_type.

    set context_id [ad_conn package_id]

    set result [db_list_of_lists km_linked_object_types "
	select ot.object_type_id, ot.pretty_name, ot.pretty_plural, ot.graphic
        from sn_questions q, sn_question_object_type_map qm, sn_object_types ot
        where q.question_id = qm.question_id
	    and qm.object_type_id = :object_type_id
	    and qm.question_state <> 'invisible'
            and ot.object_type_id = :object_type_id
            and context_id=:context_id
            and q.abstract_data_type = 'object_link'"]

    return $result
}

ad_proc km_get_linked_objects { 
    {
	-question_id 0
	-to_p 0
	-from_p 0
	-reference_links_p 1
        -family_members_p 1
	-only_unarchived_p 0
    }   
    user_id object_id
} {

    Returns objects linked to this object as specified in the
    parameters. Note that this function does not enforce
    access control rules, but only returns two colunms named
    view_p and edit_p.

} {
    if { !$to_p && !$from_p } { return {} }

    set admin_p [km_conn admin_p]
    set package_id [ad_conn package_id]
    set instance_read_p [ad_permission_p $package_id "read"]

    set join_tables {
	"sn_links l" 
	"sn_objects obj1" 
	"sn_objects obj2"
	"users u"
	"sn_object_types ot"
	"sn_question_object_type_map qotm"
	"sn_question_link_map qlm"
    }

    set order_by "order by linked_object_type, question_id, name"

    set parent_link_type parent_child

    set join_tables_from {}
    lappend where_from \
	"obj1.object_id = l.object_id_a " \
	"l.object_id_b = obj2.object_id " \
        "obj2.original_author_id = u.user_id" \
        "obj1.object_id = :object_id " \
	"obj2.expiration_date > sysdate" \
	"qotm.object_type_id = obj1.object_type_id" \
	"qotm.question_id = qlm.question_id" \
	"l.link_id = qlm.link_id" \
	"qotm.question_state <> 'invisible'" \
	"ot.object_type_id = obj2.object_type_id" \
	"ot.deleted_p = 'f'"

    if !$admin_p { lappend where_from "ot.public_p = 't'" }

    if !$instance_read_p {
	lappend where_from \
	    "(obj2.public_p = 't' or obj2.object_id in (
	      select object_id from acs_permissions 
	      where (grantee_id = :user_id
		     or grantee_id in (select group_id from membership_rels
				       where person_id = :user_id))
		  and privilege = 'read'))"
    }

    lappend where_to \
	"obj1.object_id = l.object_id_a " \
	"l.object_id_b = obj2.object_id " \
        "obj1.original_author_id = u.user_id" \
        "obj2.object_id = :object_id " \
        "obj1.expiration_date > sysdate" \
	"qotm.object_type_id = obj1.object_type_id" \
	"qotm.question_id = qlm.question_id" \
	"l.link_id = qlm.link_id" \
	"qotm.question_state <> 'invisible'" \
	"ot.object_type_id = obj1.object_type_id" \
	"ot.deleted_p = 'f'"

    if !$family_members_p {
	lappend where_to "l.link_type not in ('uncle_nephew','parent_child')"
	lappend where_from "l.link_type not in ('uncle_nephew','parent_child')"
    }

    if { $only_unarchived_p } {
	lappend where_to "obj1.archived_p = 'f'"
	lappend where_from "obj2.archived_p = 'f'"
    }

    if !$admin_p { lappend where_to "ot.public_p = 't'" }

    if !$instance_read_p {
	lappend where_to \
	    "(obj1.public_p = 't' or obj1.object_id in (
	      select object_id from acs_permissions 
	      where (grantee_id = :user_id
		     or grantee_id in (select group_id from membership_rels
				       where person_id = :user_id))
		  and privilege = 'read'))"
    }

    if { $question_id } {
	lappend where_from "qlm.question_id = :question_id"
	lappend where_to "qlm.question_id = :question_id"
    }

    if { $reference_links_p == 0 } {
	# We were told to exclude links which may already have been
	# displayed as answers in the target object.

        lappend where_to {
	    qlm.question_id not in
	    (select references_question_id
	     from sn_questions
	     where references_question_id is not null and question_id in
	     (select question_id from sn_question_object_type_map 
	      where object_type_id = obj2.object_type_id))
	}
    }

    set links_from "
	select l.link_id, l.link_comment, decode(html_p,'t',1,0) as comment_html_p,
	    obj2.object_id linked_object_id, obj2.object_type_id linked_object_type,
	    qlm.question_id, obj2.one_line_description name, obj2.last_modified,
	    km_utilities.strip_html_and_limit_size(obj2.overview) as overview,
	    decode(obj2.public_p,'t',1,0) as public_p, u.user_id as owner_id, 
	    u.first_names || ' ' || u.last_name as owner_name,
	    1 as view_p, decode(obj2.archived_p,'t',1,0) as archived_p,
            decode(obj2.in_review_p,'t',1,0) as review_p,
	    decode(acs_permission.permission_p(obj2.object_id, :user_id, 'write'),'t',1,0) as edit_p,
	    1 as from_p
        from [join [concat $join_tables $join_tables_from] ","]
        where [join $where_from " and "]"

    set links_to "
	select l.link_id, l.link_comment, decode(html_p,'t',1,0) as comment_html_p,
	    obj1.object_id linked_object_id, obj1.object_type_id linked_object_type,
	    qlm.question_id, obj1.one_line_description name, obj1.last_modified,
	    km_utilities.strip_html_and_limit_size(obj1.overview) as overview,
	    decode(obj1.public_p,'t',1,0) as public_p, u.user_id as owner_id,
	    u.first_names || ' ' || u.last_name as owner_name,
	    1 as view_p, decode(obj1.archived_p,'t',1,0) as archived_p,
            decode(obj1.in_review_p,'t',1,0) as review_p,
	    decode(acs_permission.permission_p(obj1.object_id, :user_id, 'write'),'t',1,0) as edit_p,
	    0 as from_p
	from [join $join_tables ","]
        where [join $where_to " and "]"

    # Put the SQL together.
    if { $from_p && $to_p } { 
	set sql "$links_from union $links_to $order_by"
    } elseif { $from_p } {
	set sql "$links_from $order_by"
    } else {
	set sql "$links_to $order_by"
    }

    # strip html from overview
    set result [km_db_to_named_list_list km_get_linked_objects_2 $sql]    

    return $result
}

ad_proc km8_get_to_links { {-only_unarchived_p 0} object_id question_id } {

} {
    set current_user_id [ad_conn user_id]
    set result [db_list_of_lists km8_get_to_links {
        select * from (
        select object_id_b, aodesc.shortname, aodesc.overview,
        aodesc.overview_html_p, l.link_id, object_id_a, ao.object_type,
        aodesc.presentation, acs_permission.permission_p(l.object_id_b, :current_user_id, 'sw_read') sw_read_p, 
        (select decode(count(*),0,'f','t')
	 from membership_rels mr_lm
	 where mr_lm.person_id = l.object_id_b
	 and sn_group.group_list_members_p(mr_lm.group_id, :current_user_id) > 0 ) as user_list_p,
	l.link_comment, decode(l.html_p,'t',1,0) as comment_html_p
        from sn_links l, sn_question_link_map map, acs_objects_description aodesc,
        acs_objects ao
        where l.object_id_a = :object_id
        and aodesc.object_id = ao.object_id
        and aodesc.object_id = l.object_id_b
	and aodesc.state <> 'd'
	and (:only_unarchived_p = 0 or ao.object_type <> 'sn_object' or aodesc.state <> 'a')
        and map.question_id = :question_id
        and l.link_id = map.link_id
        order by shortname
        ) where sw_read_p='t' or user_list_p='t'}]
    return $result
}

ad_proc km_get_linked_names { object_id question_id } {
    Returns the names of users, objects or ACS content linked from this object and question.
} {
    set context_id [ad_conn package_id]

    set result [db_list_of_lists km_get_linked_names "
	select obj1.one_line_description
        from sn_links l, sn_objects obj1, sn_objects obj2, sn_question_link_map map
        where l.object_id_a = obj2.object_id
            and l.object_id_b = obj1.object_id
            and map.question_id = :question_id
            and l.link_id = map.link_id
            and obj2.object_id = :object_id
            and obj1.context_id = :context_id
            and obj2.context_id = :context_id
	order by one_line_description"]

    return [map fst $result]
}

ad_proc sn_people_responsible {user_id object_id} {

    Returns a list of triples:user_id, full name, email of people
    responsible for this xchange_task project.

} {
    set package_id [ad_conn package_id]
    set contact_person_p [db_0or1row get_contact_person_question {
	select q.question_id
	from sn_objects o, sn_questions q, sn_question_object_type_map qm
        where q.question_id = qm.question_id
	    and qm.question_state = 'active'
	    and qm.object_type_id = o.object_type_id
	    and o.object_id = :object_id
	    and o.context_id = :package_id
	    and q.pretty_name = 'Contact Person'
    }]

    set result ""
    if {$contact_person_p} {
	db_0or1row get_contact_person {
	    select content as result
	    from sn_content
	    where object_id = :object_id
	    and question_id = :question_id
	}
    }
    if [empty_string_p $result] {
        db_1row get_author_email {
	    select u.first_names || ' ' || u.last_name author_name, u.email author_email
	    from users u, sn_objects o
	    where u.user_id = o.original_author_id
	    and o.object_id = :object_id
	}
	set result "$author_name <$author_email>"
    }

    return $result
}

ad_proc km_get_parent { object_id } {
    Returns the parent of an object
} {
   if { [db_0or1row km_get_parent {
	select obj.object_id as parent_id, obj.one_line_description as parent_name, obj.object_type_id
	from sn_objects obj, km_flat_object_hierarchy fo
        where fo.child = :object_id
	and obj.object_id = fo.parent
	and fo.link_type = 'parent_child'
        and fo.distance = 1 }] } {

       if {[empty_string_p $parent_name]} {
	   set parent_name "Unnamed [km_static object_type_pretty_name $object_type_id]"
       }
       return [list $parent_id $parent_name $object_type_id]
   } else {
       return
   }
}

ad_proc km_get_uncle { object_id } {
    Returns the uncle of an object
} {
    if { [db_0or1row km_get_uncle {
	select obj.object_id as uncle_id, obj.one_line_description as uncle_name, obj.object_type_id
	from sn_objects obj, km_flat_object_hierarchy fo
        where fo.child = :object_id
	and obj.object_id = fo.parent
	and fo.link_type = 'uncle_nephew'
	and fo.distance = 1 }] } {

       if {[empty_string_p $uncle_name]} {
	   set uncle_name "Unnamed [km_static object_type_pretty_name $object_type_id]"
       }
       return [list $uncle_id $uncle_name $object_type_id]
   } else {
       return
   }
}

ad_proc km_is_nephew_p { object_id } {
    Checks if this object is linked as a nephew to an uncle
} {
    set result [db_string km_is_nephew_p {
	select count(*)
	from km_flat_object_hierarchy
	where child = :object_id
	and link_type = 'uncle_nephew'
	and distance = 1
    }]
    return $result
}

ad_proc km_is_uncle_p { object_id } {
    Checks if this object is linked as a uncle to an object
} {
    set result [db_string km_is_uncle_p {
	select count(*)
	from km_flat_object_hierarchy foh, sn_objects o
	where foh.parent = :object_id
	and foh.link_type = 'uncle_nephew'
	and foh.distance = 1
	and foh.child = o.object_id
	and o.expiration_date > sysdate
    }]
    return $result
}

ad_proc km_is_parent_p { object_id } {
    Checks if this object is linked as a parent to an object
} {
    set result [db_string km_is_parent_p {
	select count(*)
	from km_flat_object_hierarchy foh, sn_objects o
	where foh.parent = :object_id
	and foh.link_type = 'parent_child'
	and foh.distance = 1
	and foh.child = o.object_id
	and o.expiration_date > sysdate
    }]
    return $result
}

ad_proc km_is_child_p { object_id } {
    Checks if this object is linked as a child to a parent
} {
    set result [db_string km_is_child_p {
	select count(*)
	from km_flat_object_hierarchy
	where child = :object_id
	and link_type = 'parent_child'
	and distance = 1
    }]
    return $result
}

ad_proc km_is_descendant_p { object_id } {
    Checks if this object is linked as a child/nephew to a parent/uncle
} {
    set result [db_string km_is_child_p {
	select count(*)
	from km_flat_object_hierarchy
	where child = :object_id
	and distance = 1
    }]
    return $result
}

ad_proc km_is_mandatory_nephew_p { object_id } {
    Checks if this object is linked as a nephew to an uncle via a mandatory question
} {
    set result [db_string km_is_mandatory_nephew_p {
	select count(*)
	from sn_links l, sn_question_link_map map, sn_question_object_type_map qm
	where l.object_id_b = :object_id
	and l.link_type = 'uncle_nephew'
	and l.link_id = map.link_id
	and map.question_id = qm.question_id
	and qm.question_state = 'active'
	and qm.mandatory_p = 't'
    }]
    return $result
}

ad_proc km_is_mandatory_uncle_p { object_id } {
    Checks if this object is linked as a uncle to an object via a mandatory question
} {
    set result [db_string km_is_mandatory_uncle_p {
	select count(*)
	from sn_links l, sn_question_link_map map,
	     sn_question_object_type_map qm, sn_objects o
	where l.object_id_a = :object_id
	and l.object_id_b = o.object_id
	and o.expiration_date > sysdate
	and l.link_type = 'uncle_nephew'
	and l.link_id = map.link_id
	and map.question_id = qm.question_id
	and qm.question_state = 'active'
	and qm.mandatory_p = 't'
    }]
    return $result
}

ad_proc km_is_nephew_with_private_uncle_p { object_id {public_p "f"} } {
    Checks if this object is linked as a nephew to an uncle
} {
    set result [db_string km_is_nephew_with_private_uncle_p {
	select count(*)
	from km_flat_object_hierarchy ko, sn_objects o
	where ko.child = :object_id
	and ko.parent = o.object_id
	and ko.link_type = 'uncle_nephew'
	and ko.distance = 1
	and o.public_p = :public_p
    }]
    return $result
}

ad_proc km_is_child_with_private_parent_p { object_id {public_p "f"} } {
    Checks if this object is linked as a child to a parent
} {
    set result [db_string km_is_child_with_private_parent_p {
	select count(*)
	from km_flat_object_hierarchy ko, sn_objects o
	where ko.child = :object_id
	and ko.parent = o.object_id
	and ko.link_type = 'parent_child'
	and ko.distance = 1
	and o.public_p = :public_p
    }]
    return $result
}

ad_proc km_private_ancestor_p { object_id {public_p "f"} } {
    Checks if this object is linked as a child or nephew to a private ancestor.
} {
    set result [db_string km_private_ancestor_p {
	select count(*)
	from km_flat_object_hierarchy ko, sn_objects o
	where ko.child = :object_id
	and ko.parent = o.object_id
	and ko.distance = 1
	and o.public_p = :public_p
    }]
    return $result
}

ad_proc km_archived_ancestor_p { object_id {archived_p "t"} } {
    Checks if this object is linked as a child or nephew to an archived ancestor.
} {
    set result [db_string km_archived_ancestor_p {
	select count(*)
	from km_flat_object_hierarchy ko, sn_objects o
	where ko.child = :object_id
	and ko.parent = o.object_id
	and ko.distance = 1
	and o.archived_p = :archived_p
    }]
    return $result
}

ad_proc km_check_public_status_after_edit { object_id } {
    Checks the public status of the object, it's descendants and it's uncle after
    editing the object.
} {
    set nephew_p [km_is_mandatory_nephew_p $object_id]
    set complete_p [km_object_complete_p $object_id]
    db_1row km_check_public_status_after_edit_10 {
	select public_p, archived_p
	from sn_objects where
	object_id = :object_id
    }
    if { $complete_p && $public_p == "f" } {
	if {[km_private_ancestor_p $object_id "t"]} {
	    # if object is uncle then publish it and all nephews - user can later
	    # unpublish some nephews
	    km_object_publish [concat $object_id [map fst [km_get_descendents $object_id]]]
	}
    }
    if { !$complete_p && $public_p == "t" } {
	km_object_unpublish $object_id
	if {$nephew_p} {
	    set uncle_id [km_get_uncle $object_id]
	    if {[km_uncle_needs_private_status_p $uncle_id]} {
		km_object_unpublish $uncle_id
	    }
	}
    }
}

ad_proc km_object_publish { object_list } {
    Publishes an object and all its selected descendants
} {
    set user_id [ad_conn user_id]

    foreach child_id $object_list {
	if [km_object_complete_p $child_id] {
	    lappend object_list $child_id
	}
    }
    set object_sql [join $object_list ","]

    db_transaction {
	db_exec_plsql km_object_publish_10 "
	    declare
	        v_the_public acs_objects.object_id%TYPE;
	    begin
                select acs.magic_object_id('the_public') into v_the_public from dual;
                for v_objects in (select object_id from sn_objects where object_id in ($object_sql))
                loop
                    acs_permission.grant_permission(v_objects.object_id, v_the_public, 'read');
	            insert into sn_audit_table
	            (object_id, question_id, last_modified, last_modifying_user_id, content)
	            values
	            (v_objects.object_id, null, sysdate, :user_id, 'Object Published');
                end loop;
	    end;
	"

	db_dml km_object_publish_20 "
	    update sn_objects
	    set user_checkoff_date = sysdate, public_p = 't', in_review_p = 'f'
	    where object_id in ($object_sql)"

    }
}

ad_proc km_object_unpublish { object_id } {
    Unpublishes an object and all its descendants
} {
    if {[ad_conn -connected_p]} {
	set user_id [ad_conn user_id]
    } else {
	set user_id [db_null]
    }

    db_transaction {
	db_exec_plsql km_object_unpublish_10 {
	    declare
	        v_the_public acs_objects.object_id%TYPE;
	    begin
	        select acs.magic_object_id('the_public') into v_the_public from dual;
                for v_objects in (select object_id from sn_objects 
                    where object_id = :object_id
	            or object_id in (select o.object_id
                            from km_flat_object_hierarchy foh, sn_objects o
                            where foh.parent = :object_id
                            and foh.child = o.object_id
                            and o.public_p = 't') )
                loop
                    acs_permission.revoke_permission(v_objects.object_id, v_the_public, 'read');
	            insert into sn_audit_table
	            (object_id, question_id, last_modified, last_modifying_user_id, content)
	            values
	            (v_objects.object_id, null, sysdate, :user_id, 'Set to Private');
                end loop;
	    end;
	}

	db_dml km_object_unpublish_20 {
	    update sn_objects
	    set user_checkoff_date = null, public_p = 'f'
	    where object_id = :object_id
	    or object_id in (select o.object_id
	                from km_flat_object_hierarchy foh, sn_objects o
                        where foh.parent = :object_id
                        and foh.child = o.object_id
                        and o.public_p = 't')
	}
    }
}

ad_proc km_object_archive { object_id reason } {
    archives an object and all its descendants
} {
    set reason "Archived: [ad_quotehtml $reason]"
    if {[ad_conn -connected_p]} {
	set user_id [ad_conn user_id]
    } else {
	set user_id [db_null]
    }

    db_transaction {
	set object_list [db_list get_archivable_objects {
	    ( select o.object_id
	    from km_flat_object_hierarchy foh, sn_objects o, sn_object_types ot
	    where foh.parent = :object_id
	    and foh.child = o.object_id
	    and o.archived_p = 'f'
	    and o.object_type_id = ot.object_type_id
	    and ot.archive_p = 't' )
	    minus
	    ( select child as object_id
	    from km_flat_object_hierarchy
	    where parent in (select o.object_id
	                     from km_flat_object_hierarchy foh, sn_objects o,
	                          sn_object_types ot
	                     where foh.parent = :object_id
	                     and foh.child = o.object_id
	                     and o.archived_p = 'f'
	                     and o.object_type_id = ot.object_type_id
	                     and ot.archive_p = 'f') )
	}]
	set object_sql [join [concat $object_id $object_list] ","]

	db_dml km_object_archive_10 "
	    insert into sn_audit_table (
	       select o.object_id, null as question_id, sysdate as last_modified,
	              :user_id as last_modifying_user_id, :reason as content
	       from sn_objects o
	       where o.object_id in ($object_sql)
            )
	"

	db_dml km_object_archive_20 "
	    update sn_objects
	    set archiving_date = sysdate,
	        archived_p = 't'
	    where object_id in ($object_sql)
	"
    }
}

ad_proc km_object_unarchive { object_id object_id_list } {
    Unarchives an object and all selected descendants (-> private)
} {
    set user_id [ad_conn user_id]
    lappend object_id_list $object_id
    set object_sql [join $object_id_list ","]

    db_transaction {
	## revoke public permissions from all objects + descendants
	db_exec_plsql km_object_unarchive_10 {
	    declare
	        v_the_public acs_objects.object_id%TYPE;
	    begin
                select acs.magic_object_id('the_public') into v_the_public from dual;
                for v_objects in (select object_id from sn_objects 
                    where object_id = :object_id
	            or object_id in (select o.object_id
                            from km_flat_object_hierarchy foh, sn_objects o
                            where foh.parent = :object_id
                            and foh.child = o.object_id
                            and o.public_p = 't') )
	        loop
	            acs_permission.revoke_permission(v_objects.object_id, v_the_public, 'read');
	        end loop;
	    end;
	}

	## make a note in the audit table that objects got unarchived
	db_dml km_object_unarchive_20 "
	    insert into sn_audit_table (
	       select o.object_id, null as question_id, sysdate as last_modified,
	              :user_id as last_modifying_user_id, 'Unarchived, set to Private' as content
	       from sn_objects o
	       where object_id in ($object_sql)
            )
	"

	## remove objects from archive and make private
	db_dml km_object_unarchive_30 "
	    update sn_objects
	    set archiving_date = null,
	        user_checkoff_date = null,
	        public_p = 'f',
	        archived_p = 'f'
	    where object_id in ($object_sql)
	"

	## make a note in the audit table that all other descendants
	## got private
	db_dml km_object_unarchive_40 "
	    insert into sn_audit_table (
	       select o.object_id, null as question_id, sysdate as last_modified,
	              :user_id as last_modifying_user_id, 'Set to Private' as content
	       from km_flat_object_hierarchy foh, sn_objects o
	       where foh.parent = :object_id
	       and foh.child = o.object_id
	       and o.public_p = 't'
            )
	"

	## make all other desendants private
	db_dml km_object_unarchive_50 "
	    update sn_objects
	    set user_checkoff_date = null,
	        public_p = 'f'
	    where object_id in (select o.object_id
	                from km_flat_object_hierarchy foh, sn_objects o
	                where foh.parent = :object_id
	                and foh.child = o.object_id
	                and o.public_p = 't')
	"
    }
}

ad_proc km_get_object_ancestors { object_id } {

    Returns a list of ids of the parent of this object, the parent of 
    the parent of this object, etc. Will do this for normal links if
    the link type_id is set to the one for bidirectional links.

} {
    set result [db_list_of_lists km_get_object_ancestors "
	select obj.object_id, obj.one_line_description, obj.object_type_id
	from sn_objects obj, km_flat_object_hierarchy fo
        where fo.child = :object_id
	    and obj.object_id = fo.parent
	order by distance"]

    return $result
}

ad_proc km_get_descendents { object_id } {
    Almost identical to the ancestor proc.
} {
    set result [db_list_of_lists km_get_descendents "
	select obj.object_id, obj.one_line_description
	from sn_objects obj, km_flat_object_hierarchy fo
        where fo.parent = :object_id
        and obj.object_id = fo.child
        and obj.expiration_date > sysdate
	order by distance"]

    return $result
}

ad_proc km_get_child_descendents { object_id } {
    Returns only children and grandchildren.
} {
    set result [db_list km_get_child_descendents {
	select fo.child
	from km_flat_object_hierarchy fo, sn_objects o
        where fo.parent = :object_id
	and fo.link_type = 'parent_child'
	and o.object_id = fo.child
	and o.expiration_date > sysdate
	order by distance }]

    return $result
}

ad_proc km_get_objects_descendents { object_list } {
    Returns descendents of a list of objects.
} {
    set result [db_list km_get_objects_descendents "
	select fo.child
	from km_flat_object_hierarchy fo, sn_objects o
        where fo.parent in ([join $object_list ","])
        and o.object_id = fo.child
        and o.expiration_date > sysdate
	order by distance"]

    return $result
}

ad_proc km_get_presentation_type_of_nephew_question { object_id } {
    Returns the presentation_type of the uncle's nephew question under
    which this object is linked
} {
    set result [db_string km_get_presentation_type_of_nephew_question_10 {
	select q.presentation_type
	from sn_questions q, sn_question_link_map map, sn_links l
	where q.abstract_data_type = 'nephew_object'
	and q.question_id = map.question_id
	and l.link_id = map.link_id
	and l.link_type = 'uncle_nephew'
	and l.object_id_b = :object_id
    } -default "custom"]

    return $result
}

ad_proc km_get_child_objects { 
    {-only_unarchived_p 0} object_id question_id 
} {
    Returns an array containing the object_ids of the child objects
    associated with this object_id and question_id.

} {
    set order_by [db_string order_by {
	select nvl(order_by,'name')
	from sn_questions
	where question_id = :question_id}]

    # We select more columns than we need because of sorting.
    return [map fst [db_list child_objects {
	select obj2.object_id, obj2.one_line_description as name, obj2.last_modified, l.link_id
	from sn_objects obj1,sn_objects obj2, sn_links l, sn_question_link_map map
        where l.object_id_b = obj2.object_id
	    and l.object_id_a = obj1.object_id
            and map.question_id = :question_id
            and l.link_id = map.link_id
            and l.link_type = 'parent_child'
            and obj1.object_id = :object_id
            and obj2.expiration_date > sysdate
	    and obj2.archived_p <> decode(:only_unarchived_p,'0','a','t')
	order by :order_by}]]
}

ad_proc km_get_nephew_objects { 
    {-only_unarchived_p 0} object_id question_id 
} {
    Returns an array containing the object_ids of the nephew objects
    associated with this object_id and question_id.

} {
    set user_id [ad_conn user_id]
    set order_by [db_string order_by {
	select nvl(order_by,'name')
	from sn_questions
	where question_id = :question_id}]

    # We select more columns than we need because of sorting.
    return [map fst [db_list nephew_objects {
	select obj2.object_id, obj2.one_line_description as name, obj2.last_modified, l.link_id
	from sn_objects obj1,sn_objects obj2, sn_links l, sn_question_link_map map
        where l.object_id_b = obj2.object_id
	    and l.object_id_a = obj1.object_id
            and map.question_id = :question_id
            and l.link_id = map.link_id
            and l.link_type = 'uncle_nephew'
            and obj1.object_id = :object_id
            and obj2.expiration_date > sysdate
	    and obj2.archived_p <> decode(:only_unarchived_p,'0','a','t')
	    and acs_permission.permission_p (obj2.object_id, :user_id, 'read') = 't'
	order by :order_by}]]
}

ad_proc km_get_nephew_count { object_id } {
    Returns the total number of nephew objects for a particular object.
} {
    set result [db_string km_get_nephew_count {
	select count(*)
	from sn_links l, sn_objects o
	where l.link_type = 'uncle_nephew'
	and l.object_id_a = :object_id
	and l.object_id_b = o.object_id
	and o.expiration_date > sysdate
    }
    return $result
}

ad_proc km_uncle_needs_private_status_p { object_id {nephew_id ""} } {
    Checks if uncle need to be private after making a nephew private
} {
    set context_id [ad_conn package_id]
    if {[empty_string_p $nephew_id]} {
	set nephew_questions_without_public_nephews [db_string km_uncle_needs_private_status_p_10 {
	select count(*)
        from sn_questions q, sn_question_object_type_map qm, sn_objects o
        where q.question_id = qm.question_id
            and qm.question_state = 'active' 
	    and qm.mandatory_p = 't'
	    and qm.object_type_id = o.object_type_id
            and o.object_id = :object_id
            and o.context_id=:context_id
	    and q.abstract_data_type = 'nephew_object'
	    and not exists (select nephew.object_id
	                    from sn_objects nephew, sn_question_link_map map, sn_links l
	                    where l.object_id_a = :object_id
	                    and l.object_id_b = nephew.object_id
	                    and l.link_type = 'uncle_nephew'
	                    and l.link_id = map.link_id
	                    and map.question_id = q.question_id
	                    and nephew.expiration_date > sysdate
	                    and nephew.archived_p = o.archived_p
	                    and nephew.public_p = 't')
	}]
    } else {
	set nephew_questions_without_public_nephews [db_string km_uncle_needs_private_status_p_20 {
	select count(*)
        from sn_questions q, sn_question_object_type_map qm, sn_objects o
        where q.question_id = qm.question_id
            and qm.question_state = 'active' 
	    and qm.mandatory_p = 't'
	    and qm.object_type_id = o.object_type_id
            and o.object_id = :object_id
            and o.context_id=:context_id
	    and q.abstract_data_type = 'nephew_object'
	    and not exists (select nephew.object_id
	                    from sn_objects nephew, sn_question_link_map map, sn_links l
	                    where l.object_id_a = :object_id
	                    and l.object_id_b = nephew.object_id
	                    and l.link_type = 'uncle_nephew'
	                    and l.link_id = map.link_id
	                    and map.question_id = q.question_id
	                    and nephew.expiration_date > sysdate
	                    and nephew.public_p = 't'
	                    and nephew.archived_p = o.archived_p
	                    and nephew.object_id <> :nephew_id)
	}]
    }
    return $nephew_questions_without_public_nephews
}

ad_proc km_link_question_attributes { question_id } {
    Returns the object_type_attributes for this object_link question.
} {
    set context_id [ad_conn package_id]
    set sql "select ot.object_type_id, ot.pretty_name, ot.pretty_plural, ot.graphic 
             from sn_questions q, sn_object_types ot 
             where q.target_object_type_id = ot.object_type_id
             and q.question_id = :question_id"
    set result [km_db_1row_to_list km_link_question_attributes $sql]
    return $result
}

ad_proc km_link_objects { 
    {
        -link_type ""
    } 
    question_id user_id object_id_a object_id_b link_comment html_p
} {
    Creates a new link between two objects.
} {
    if [empty_string_p $link_type] {
        set link_type bi_directional
    }

    if {[string equal $link_type "parent_child"] || [string equal $link_type "uncle_nephew"]} {
	if {[db_string check_parent_or_uncle "
	select count(*) from sn_links
	where object_id_b = :object_id_b
	and link_type in ('parent_child','uncle_nephew')"] > 0} {
	    return 0
	}
    }

    set new_link_id [db_string km_link_objects_10 "select sn_links_seq.nextval from dual"]

    db_dml km_link_objects_20 {insert into sn_links (link_id, link_type, object_id_a,
    object_id_b, link_comment, html_p, creation_user, creation_date)
    values (:new_link_id, :link_type, :object_id_a,
    :object_id_b, :link_comment, :html_p, :user_id, sysdate)}

    if { ![db_0or1row km_link_objects_25 {
	select nvl(one_line_description,'Unnamed ' || pretty_name) as target_name, pretty_name as target_type
	from sn_objects o, sn_object_types ot
	where o.object_id = :object_id_b
	  and o.object_type_id = ot.object_type_id}]
    } {
	db_1row km_link_objects_26 {
	    select 
              shortname as target_type,
              overview  as target_name
	    from
	      acs_objects_description
	    where
	      object_id = :object_id_b
	}
    }
           
    set content "Added link to $target_type \"$target_name\" $link_comment"
    set ip_address [ns_conn peeraddr]

    db_exec_plsql km_link_objects_30 {
	begin
	object.audit_object (
	v_object_id              => :object_id_a,
	v_question_id            => :question_id,
	v_last_modifying_user_id => :user_id,
	v_modifying_ip           => :ip_address,
	v_content                => :content
	);
	end;
    }

    db_dml km_link_objects_40 {insert into sn_question_link_map (link_id, question_id)
    values (:new_link_id, :question_id)}

    set sql_fragment "[km_link_description_update $object_id_a $question_id]"
    if {![empty_string_p $sql_fragment]} {
        db_dml km_link_objects_50 "
        update sn_objects 
        set $sql_fragment
        where object_id = :object_id_a"
    }
    return 1
}

ad_proc km_link_description_update { object_id question_id } {

    Returns a SQL fragment for updating the short or long description
    of an object, if the given question is set as the object type's
    description. Works for all linking questions.

} {
    db_1row km_link_description_update "
	select long_description
	from sn_object_types
	where object_type_id = (select object_type_id from sn_objects 
				where object_id=:object_id)"

    set short_desc_list [db_list_of_lists question_info_short_desc {
	select map.short_description
        from sn_types_map_short_name map
        where map.object_type_id = (select object_type_id from sn_objects 
				where object_id=:object_id)
        order by position
    }]

    set result ""
    if { [lsearch $short_desc_list $question_id] != -1 || $question_id == $long_description } {
	set QQlinked_names [util_trim_string_with_hrefs [DoubleApos [join [km_get_linked_names $object_id $question_id] ", "]] 2000]

	if { $question_id == $long_description } {
	    set result "overview = '$QQlinked_names', overview_html_p = 'f'"
	} else {
	    set result "one_line_description = '$QQlinked_names'"
	}
    }

    return $result
}

namespace eval sn_links::objects::sn_object {
    ad_proc -public restrict {source_id target_id target_presentation pass} {

	Limit knowledge object types in linking service.

    } {

	if { [regexp {exclude ([0-9 ]+)} $pass match exclude_ids] } {
	    if { [lsearch $exclude_ids $target_id] != -1 } {
		return 0
	    }
	}

	if {![regexp {restriction(\d*)} $pass match restrict_to_object_type_id]} {
	    # The restriction is not required, so show all objects.
	    return 1
	} else {
	    if {![regexp {object_type_id=(\d*)} $target_presentation match object_type_id]} {
	    # Bug in the database, the presentation column is wrong.
		ns_log Notice "Presentation is wrong for object_id $target_id."
		return 1
	    } else {
		if {$object_type_id == $restrict_to_object_type_id} {
		    return 1
		}
		return 0
	    }
	}
    }
}
